import collections
import sys
import hashlib
import re
import itertools
import os

scrabble = {'E':1, 'A':1, 'I':1, 'O':1, 'N':1, 'R':1, 'T':1, 'L':1, 'S':1, 'U':1,
            'D':2, 'G':2, 'B':3, 'C':3, 'M':3, 'P':3,
            'F':4, 'H':4, 'V':4, 'W':4, 'Y':4, 'K':5,
            'J':8, 'X':8, 'Q':10, 'Z':10}

with open("words.txt") as fp:
    words = set(fp.read().splitlines())

def maxfreq(w, filter=None):
    if filter: filter = set(filter)
    freq = collections.Counter()
    for c in w:
        if filter is None or c in filter:
            freq[c] += 1
    try:
        return max(freq.items(), key=lambda (chr, frq): frq)[1]
    except ValueError:
        return 0

def check_complete(*sets):
    freqs = collections.Counter()
    for s in sets:
        for v in s:
            freqs[v] += 1
    assert len(freqs) == 26
    assert all(v == 1 for v in freqs.itervalues())

vowels = "aeiou"
consonants = "bcdfghjklmnpqrstvwxyz"
check_complete(vowels, consonants)

toprow = "qwertyuiop"
midrow = "asdfghjkl"
botrow = "zxcvbnm"
check_complete(toprow, midrow, botrow)

class LookupCache(object):
    def __init__(self, fn):
        self.fn = fn
        self.cache = {}
    def __call__(self, word):
        if word not in self.cache:
            self.cache[word] = self.fn(word)
        return self.cache[word]

if not os.path.exists("tiles.pickle"):
    import tile
    tile.main()

with open("tiles.pickle") as fp:
    import cPickle
    cached_tile_info = cPickle.load(fp)

def caesar_exists_impl(word):
    def caesar(w, sh):
        return "".join(chr(ord('a') + (ord(c) - ord('a') + sh)%26) for c in w)
    for sh in xrange(1, 26):
        if caesar(word, sh) in words:
            return True
    return False
caesar_exists = LookupCache(caesar_exists_impl)

anagram_freq = collections.Counter()
for w in words:
    anagram_freq["".join(sorted(w))] += 1

def anagram_exists_impl_for(extrachars):
    def anagram_exists_impl(word):
        for addend in itertools.product(*(vowels + consonants
                                          for i in xrange(extrachars))):
            freq = anagram_freq["".join(sorted(word + "".join(addend)))]
            if freq >= 2 or (freq == 1 and (extrachars > 0 or
                                            word not in words)):
                return True
        return False
    return anagram_exists_impl

anagram_exists_for = {2: LookupCache(anagram_exists_impl_for(2)),
                      1: LookupCache(anagram_exists_impl_for(1)),
                      0: LookupCache(anagram_exists_impl_for(0))}

def word_to_base26_number(word):
    res = 0
    for c in word:
        res *= 26
        res += ord(c) - ord('a')
    return res

def is_float_representable(n, mantbits):
    # trailing zeroes are fine as long as we don't overflow the exponent
    # (which won't happen for any words; 'antidisestablishmentarianism'
    # maps to a 125-bit number, and a single precision float can go up to
    # exponent 128).
    while n and (n & 1 == 0):
        n >>= 1

    # Because of the implicit leading 1., a 52-bit mantissa can represent
    # values up to 53 bits long.
    return n < 2**(mantbits+1)

is_floatrep_word = LookupCache(
    lambda w: is_float_representable(word_to_base26_number(w), 23))
is_doublerep_word = LookupCache(
    lambda w: is_float_representable(word_to_base26_number(w), 52))
is_32brep_word = LookupCache(
    lambda w: word_to_base26_number(w) < 2**32)
is_64brep_word = LookupCache(
    lambda w: word_to_base26_number(w) < 2**64)


def divisibility_properties(transform, descr, divisors=(2,3,5,7)):
    return [("%s is divisible by %d" % (descr, div),
             lambda w, t=transform, d=div: t(w) % d == 0)
            for div in divisors]

def make_puzzle_filters(NumericHandler, ContentsHandler, ComboHandler):
    return [
        NumericHandler(lambda w: len(w),
                       "Length: %(bounds)s letter%(plural)s"),
        NumericHandler(lambda w: sum(1 for c in w if c in "aeiou"),
                       "Vowels: %(bounds)s",
                       "Vowels: %(bounds)s of the letters"),
        NumericHandler(lambda w: sum(scrabble[c] for c in w.upper()),
                       "Base Scrabble score: %(bounds)s point%(plural)s"),
        NumericHandler(lambda w: sum(ord(c) - ord('a') + 1 for c in w),
                       "Sum of letters (A=1, B=2, etc): %(bounds)s"),
        [
            NumericHandler(lambda w: maxfreq(w),
                           "Most common letter(s) each appear(s): %(bounds)s time%(plural)s",
                           "Most common letter(s) each account(s) for: %(bounds)s of the letters"),
            NumericHandler(lambda w: maxfreq(w, vowels),
                           "Most common vowel(s) each appear(s): %(bounds)s time%(plural)s",
                           "Most common vowel(s) each account(s) for: %(bounds)s of the letters"),
            NumericHandler(lambda w: maxfreq(w, consonants),
                           "Most common consonant(s) each appear(s): %(bounds)s time%(plural)s",
                           "Most common consonant(s) each account(s) for: %(bounds)s of the letters"),
        ],

        [
            NumericHandler(lambda w: len(set(w)),
                           "Distinct letters: %(bounds)s"),
            NumericHandler(lambda w: len(set(w) & set(vowels)),
                           "Distinct vowels: %(bounds)s"),
            NumericHandler(lambda w: len(set(w) & set(consonants)),
                           "Distinct consonants: %(bounds)s"),
        ],

        [
            NumericHandler(lambda w: sum(1 for c in w if c in toprow),
                           "Letters located in the top row on a QWERTY keyboard: %(bounds)s",
                           "Letters located in the top row on a QWERTY keyboard: %(bounds)s of the letters"),
            NumericHandler(lambda w: sum(1 for c in w if c in midrow),
                           "Letters located in the middle row on a QWERTY keyboard: %(bounds)s",
                           "Letters located in the middle row on a QWERTY keyboard: %(bounds)s of the letters"),
            NumericHandler(lambda w: sum(1 for c in w if c in botrow),
                           "Letters located in the bottom row on a QWERTY keyboard: %(bounds)s",
                           "Letters located in the bottom row on a QWERTY keyboard: %(bounds)s of the letters"),
        ],

        [
            NumericHandler(cached_tile_info["states"].get,
                           "If you marked nonoverlapping "
                           "US state postal abbreviations, "
                           "you could mark at most: %(bounds)s letter%(plural)s",
                           "If you marked nonoverlapping "
                           "US state postal abbreviations, "
                           "you could mark at most: %(bounds)s of the letters"),
            NumericHandler(cached_tile_info["countries"].get,
                           "If you marked nonoverlapping "
                           "officially-assigned ISO 3166-1 alpha-2 country codes, "
                           "you could mark at most: %(bounds)s letter%(plural)s",
                           "If you marked nonoverlapping "
                           "officially-assigned ISO 3166-1 alpha-2 country codes, "
                           "you could mark at most: %(bounds)s of the letters"),
            NumericHandler(cached_tile_info["elements"].get,
                           "If you marked nonoverlapping "
                           "chemical element symbols (atomic number 112 or below), "
                           "you could mark at most: %(bounds)s letter%(plural)s",
                           "If you marked nonoverlapping "
                           "chemical element symbols (atomic number 112 or below), "
                           "you could mark at most: %(bounds)s of the letters"),
            NumericHandler(cached_tile_info["wl3"].get,
                           "If you marked nonoverlapping "
                           "occurrences of words in the word list that are 3 or fewer letters long, "
                           "you could mark at most: %(bounds)s letter%(plural)s",
                           "If you marked nonoverlapping "
                           "occurrences of words in the word list that are 3 or fewer letters long, "
                           "you could mark at most: %(bounds)s of the letters"),
        ],

        ContentsHandler("%(action)s: %(what)s",
                        lambda w: w, max_infix=3, max_anchored=4),
        ContentsHandler("SHA-1 hash of lowercased word, expressed "
                        "in hexadecimal, %(action)s: %(what)s",
                        lambda w: hashlib.sha1(w).hexdigest(),
                        max_infix=4, max_anchored=4),

        ComboHandler([("Starts with a vowel",
                       lambda w: w[0] in vowels),
                      ("Contains at least one doubled letter",
                       lambda w: any(first == second for first, second
                                     in zip(w[:-1], w[1:]))),
                      ("Contains at least two nonoverlapping "
                       "occurrences of the same doubled letter",
                       lambda w: bool(re.search(r'(.)\1.*\1\1', w))),
                      ("Contains at least two different doubled letters",
                       lambda w: len(set(f for f, s in zip(w[:-1], w[1:])
                                         if f == s)) >= 2),
                      ]),

        ComboHandler([("Has at least one anagram that is also "
                       "in the word list",
                       anagram_exists_for[0]),
                      ("Can be combined with one additional "
                       "letter to produce an anagram of something in the "
                       "word list",
                       anagram_exists_for[1]),
                      ("Can be combined with two additional "
                       "letters to produce an anagram of something in the "
                       "word list",
                       anagram_exists_for[2]),
                      ("Can be Caesar shifted to "
                       "produce another word in the word list",
                       caesar_exists)]),

        ComboHandler(divisibility_properties(
                word_to_base26_number,
                "Word interpreted as a base 26 number (A=0, B=1, etc)")),
        ComboHandler(divisibility_properties(
                lambda w: sum(ord(c) - ord('a') + 1 for c in w),
                "Sum of letters (A=1, B=2, etc)")),

        ComboHandler([("Word interpreted as a base 26 number (A=0, B=1, "
                       "etc) is exactly representable in "
                       "IEEE 754 double-precision floating point format",
                       is_doublerep_word),
                      ("Word interpreted as a base 26 number (A=0, B=1, "
                       "etc) is exactly representable in "
                       "IEEE 754 single-precision floating point format",
                       is_floatrep_word),
                      ("Word interpreted as a base 26 number (A=0, B=1, "
                       "etc) is representable as "
                       "an unsigned 32-bit integer",
                       is_32brep_word),
                      ("Word interpreted as a base 26 number (A=0, B=1, "
                       "etc) is representable as "
                       "an unsigned 64-bit integer",
                       is_64brep_word)])
    ]
