#!/usr/bin/env python

import re
import math
import os
import sys
import collections

from PIL import Image

def my_escape(s):
    return re.sub(r'\\%\\\(([a-z]+)\\\)s', r'%(\1)s', re.escape(s))

class NumericProperty(object):
    PERCENT_BOUNDS = (r'(?:[Ee]xactly ([\d.]+)%|'
                      r'[Bb]etween ([\d.]+)% and ([\d.]+)% \(inclusive\))')
    VALUE_BOUNDS = (r'(?:(\d+)|between (\d+) and (\d+) \(inclusive\))')

    def __init__(self, determiner, numeric_template, percentage_template=None):
        self.determiner = determiner
        self.numeric_template = numeric_template
        self.percentage_template = percentage_template

    def matches_description(self, descr):
        if '%' in descr:
            if self.percentage_template is None:
                return None
            template = my_escape(self.percentage_template) % {
                'bounds': self.PERCENT_BOUNDS}
            kind = 'percentage'
        else:
            template = my_escape(self.numeric_template) % {
                'bounds': self.VALUE_BOUNDS, 'plural': 's?'}
            kind = 'value'

        m = re.match(template, descr)
        if m is None:
            return None

        exactly, both_low, both_high = m.groups()
        if kind == 'percentage':
            valuetype = float
        else:
            valuetype = int
        minimum = valuetype(0)
        epsilon = valuetype(0.05)

        if exactly:
            low = valuetype(exactly) - epsilon
            high = valuetype(exactly) + epsilon
        else:
            low = valuetype(both_low)
            high = valuetype(both_high)

        return (kind, low, high)

    def matches_word(self, descmatch, word):
        kind, low, high = descmatch
        if kind == 'percentage':
            low *= len(word) * 0.01
            high *= len(word) * 0.01
        return low <= self.determiner(word) <= high

class ContentsProperty(object):
    def __init__(self, template, transform, **extra):
        self.transform = transform
        self.template = template

    def matches_description(self, descr):
        template = my_escape(self.template)
        template %= {'action': '(?P<action>[Cc]ontains|[Ss]tarts with|[Ee]nds with)',
                     'plural': 's?',
                     'what': "(?P<what>.+)"}
        m = re.match(template, descr)
        if m is None:
            return None
        return (m.group("action"), m.group("what").lower())

    def matches_word(self, descmatch, word):
        action, what = descmatch
        if action.lower() == "starts with":
            return self.transform(word).startswith(what)
        elif action.lower() == "ends with":
            return self.transform(word).endswith(what)
        else:
            assert action.lower() == "contains"
            return what in self.transform(word)

class BooleanProperty(object):
    def __init__(self, tag, decider):
        self.decider = decider
        self.tag = tag

    def matches_description(self, descr):
        if descr == self.tag + ": YES":
            return (True,)
        elif descr == self.tag + ": NO":
            return (False,)
        else:
            return None

    def matches_word(self, descmatch, word):
        expected, = descmatch
        return self.decider(word) == expected

from puzzlefilters import make_puzzle_filters
puzzle_filters = make_puzzle_filters(NumericProperty, ContentsProperty,
                                     lambda combo: [BooleanProperty(*prop)
                                                    for prop in combo])

class CantSolveYet(Exception): pass
kwrule_meanings = {}

def parse_filter(filter_text):
    m = re.match(r'Has property ([A-Z]+): YES$', filter_text)
    if m is not None:
        if kwrule_meanings:
            return kwrule_meanings[m.group(1).lower()]
        raise CantSolveYet

    for handler_or_list in puzzle_filters:
        for filter_handler in (handler_or_list if isinstance(handler_or_list, list)
                               else [handler_or_list]):
            res = filter_handler.matches_description(filter_text)
            if res is not None:
                return lambda w: filter_handler.matches_word(res, w)
    raise RuntimeError("Can't match filter: %s" % filter_text)

with open("words.txt") as fp:
    words = set(fp.read().splitlines())

wordlist = list(words)

solved = {}

property_words = collections.defaultdict(lambda: ([], []))
conceptcolor_words = collections.defaultdict(list)

def solve_one(rownr, colnr, puzzletext):
    lines = puzzletext.splitlines()
    filters = []
    informational = []
    curlist = filters

    for line in lines:
        if line == "":
            curlist = informational
        else:
            curlist.append(line)

    expanded_filters = []
    for filter in filters:
        expanded_filters.append(filter)

    try:
        callable_filters = map(parse_filter, expanded_filters)
    except CantSolveYet:
        return False

    if (rownr, colnr) not in solved:
        candidates = wordlist
        for filter in callable_filters:
            candidates = [w for w in candidates if filter(w)]

        if len(candidates) == 0:
            print "\n! No answer for (%d, %d)." % (rownr, colnr)
        elif len(candidates) > 1:
            print "\n! Ambiguous answer for (%d, %d): %s." % (
                rownr, colnr, ", ".join(c.upper() for c in candidates))
        else:
            solved[(rownr, colnr)] = next(iter(candidates))

        if (rownr, colnr) not in solved:
            return False

    answer = solved[(rownr, colnr)]

    for info in informational:
        m = re.match(r'This word is associated with the (.*)\.$', info)
        if m is not None:
            conceptcolor_words[m.group(1)].append(answer.upper())
        m = re.match(r'This is (NOT )?a word with property (.*)\.$', info)
        if m is not None:
            property_words[m.group(2)][m.group(1) is None].append(answer)

    return True

def solve():
    try:
        with open("solved.pickle") as fp:
            import cPickle
            solved.update(cPickle.load(fp))
    except IOError:
        pass

    if len(solved) == 10000: return

    num_solved = 0
    num_skipped = 0
    for dirname in os.listdir("."):
        if dirname.startswith("row"):
            rownr = int(dirname[3:])
            for puzzle in os.listdir("%s" % dirname):
                m = re.match(r'%s_col(\d+).txt' % dirname, puzzle)
                if m is not None:
                    colnr = int(m.group(1))
                    with open("%s/%s_col%d.txt"
                              % (dirname, dirname, colnr)) as fp:
                        puzzletext = fp.read()
                    if solve_one(rownr, colnr, puzzletext):
                        num_solved += 1
                    else:
                        num_skipped += 1
                    sys.stdout.write("%d puzzles solved, %d skipped\r"
                                     % (num_solved, num_skipped))
                    sys.stdout.flush()

    # dump solving state to remember it across runs
    with open("solved.pickle", "w") as fp:
        import cPickle
        cPickle.dump(solved, fp)

    # dump human-readable solving state
    arr = [["" for i in xrange(142)] for j in xrange(125)]
    for (r, c), w in solved.iteritems():
        arres[r][c] = w
    with open("pyramid.json", "w") as fp:
        fp.write("[\n" + "".join("    " + repr(row).replace("'", '"') +
                                 ("," if i < len(arr)-1 else "") + "\n"
                                 for i, row in enumerate(arr)) + "]\n")

    # dump informational state for human examination to fill in the
    # dictionaries below
    with open("informational.pickle", "w") as fp:
        cPickle.dump((dict(conceptcolor_words), dict(property_words)), fp)

    print "\nAll done!"

# -------------------------------------------------------------------------
# these were deduced from the output of a solve() that did not have this info

colors = {"ahmose": "blue",
          "dahshur": "gray",
          "djedefre": "teal",
          "djedkareisesi": "silver",
          "giza": "lime",
          "illahun": "maroon",
          "khaba": "yellow",
          "khafra": "aqua",
          "khendjer": "fuchsia",
          "merikare": "green",
          "niuserre": "black",
          "sahure": "olive",
          "saqqara": "red",
          "sneferu": "navy",
          "teti": "purple",
          "userkaf": "white"}

concepts = {"senusret": "color", "khufu": "then",
            "sekhemkhet": "everything"}

properties = {"unas": "allvowels",
              "menkaure": "is_sorted",
              "bikheris": "has_sh",
              "amenyqemau": "first_equals_last",
              "djoser": "has_ch",
              "khui": "plural",
              "abusir": "startex",
              "merenre": "cv_alternating",
              "hawara": "triplecons",
              "sobekneferu": "halfvowelplus",
              "pepi": "triplevowel",
              "nuri": "first_two_equals_last_two",
              "amenemhat": "startu",
              "neferefre": "qwerty",
              "meidum": "onevowelkind",
              "mazghuna": "double",
              "sethka": "edoring",
              "qakareibi": "startb",
              "lisht": "novowels",
}

def mkrules():
    def plural(w): return w.endswith("s")
    def double(w): return any(first == second for first, second in zip(w[:-1], w[1:]))
    def halfvowelplus(w): return sum(1 for c in w if c in "aeiou") >= len(w)/2.0
    def qwerty(w): return all(c in "qwertyuiop" for c in w)
    def edoring(w): return w.endswith(("ed", "ing"))
    def triplevowel(w): return any(w.count(v) >= 3 for v in "aeiou")
    def triplecons(w): return any(w.count(v) >= 3 for v in "bcdfghjklmnpqrstvwxyz")
    def startu(w): return w.startswith("u")
    def unmatched(w): return True
    def onevowelkind(w): return sum(1 for v in "aeiou" if v in w) == 1
    def allvowels(w): return sum(1 for v in "aeiou" if v in w) == 5
    def novowels(w): return sum(1 for v in "aeiou" if v in w) == 0
    def first_equals_last(w): return w[0] == w[-1]
    def first_two_equals_last_two(w): return w[0:2] == w[-2:]
    def is_sorted(w): return all(first <= second for first, second in zip(w, w[1:]))
    def has_sh(w): return "sh" in w
    def has_ch(w): return "ch" in w
    def startex(w): return w.startswith("ex")
    def startb(w): return w.startswith("b")
    def cv_alternating(w):
        evens = [c in "aeiou" for c in w[::2]]
        odds = [c in "aeiou" for c in w[1::2]]
        return (all(evens) and not any(odds)) or (all(odds) and not any(evens))
    return locals()

# to simulate solving the puzzle without knowing what the properties
# mean, comment out this line
kwrule_meanings.update({kwrule: mkrules()[meaning]
                        for kwrule, meaning in properties.iteritems()})

allkw = {}
allkw.update(properties)
allkw.update(concepts)
allkw.update(colors)

# to simulate solving the puzzle without knowing the keywords are used
# as solutions to some puzzles in the bottom row, comment out this line
wordlist.extend(allkw.keys())

def get_instructions():
    message = []
    for rowcol, word in solved.iteritems():
        if word in allkw:
            message.append((rowcol, allkw[word]))

    message.sort()
    return [instruction for rc, instruction in message]

colornames = {(0,0,0): "black", (128,128,128): "gray",
              (128,0,0): "maroon", (255,0,0): "red",
              (0,128,0): "green", (0,255,0): "lime",
              (128,128,0): "olive", (255,255,0): "yellow",
              (0,0,128): "navy", (0,0,255): "blue",
              (128,0,128): "purple", (255,0,255): "fuchsia",
              (0,128,128): "teal", (0,255,255): "aqua",
              (192,192,192): "silver", (255,255,255): "white"}
colorbyname = {n: c for c, n in colornames.iteritems()}

def follow_instructions(instructions):
    pixels = {}
    curcolor = None
    currule = None
    instructions.append("then")
    rules = mkrules()
    for instr in instructions:
        if instr == "color":
            curcolor = currule = None
        elif instr == "then":
            for rowcol, answer in solved.iteritems():
                if rules[currule](answer):
                    pixels[rowcol] = curcolor
        elif currule is None:
            currule = "unmatched" if instr == "everything" else instr
        elif curcolor is None:
            curcolor = instr
        else:
            assert False

    result = Image.new("RGBA", (max(c for r, c in solved) + 1,
                                max(r for r, c in solved) + 1),
                       (0, 0, 0, 0))
    for rowcol, colorname in pixels.iteritems():
        result.putpixel((rowcol[1], rowcol[0]),
                        colorbyname[colorname] + (255,))
    result.save("image-solved.png")

if __name__ == '__main__':
    solve()

    # to simulate trying to solve the puzzle without knowing how the
    # keywords work, don't do any of this - just solve()
    instructions = get_instructions()
    invprop = {v: k for k, v in properties.iteritems()}
    print " ".join(instructions)
    follow_instructions(instructions)
